//
//  Frame I/O and bitstream functions
//
#include "defines.h"

int depthInt9x9Sym[] = 
{
  1, 2, 3, 4, 5, 
  2, 3, 4, 5, 6, 
  3, 4, 5, 6, 7, 
  4, 5, 6, 7, 8, 
  5, 6, 7, 8, 9, 9 
};

int depthInt7x7Sym[] = 
{
  1, 2, 3, 4, 
  2, 3, 4, 5, 
  3, 4, 5, 6, 
  4, 5, 6, 7, 7 
};

int depthInt5x5Sym[] = 
{
  1, 2, 3,  
  2, 3, 4, 
  3, 4, 5, 5  
};

int *pDepthIntTab[3]={
  depthInt5x5Sym, 
  depthInt7x7Sym, 
  depthInt9x9Sym, 
};

//
//
//
int read_one_new_frame(int height, int width, int skip, int start, int imgNo, FILE *p_in, imgStruct *img)
{
  int i, j, uv;
  long status = 0; 
  int frameNo;
  int frameSize = height * width * 3 / 2;
  static byte *buf = NULL;
  byte *tmp_buf = NULL;
  static int first = 1;

  if(first)
  {
    buf = (byte *) malloc(frameSize * sizeof(byte));
    first = 0;
  }

  frameNo = imgNo * (skip + 1) + start;
   
  status |= fseek(p_in, frameNo * frameSize, SEEK_SET);
  status |= (fread(buf, frameSize * sizeof(byte), 1, p_in) != 1);

  if(!status)
  {
    tmp_buf = buf;

    for(j=0; j < height; j++)
      for(i=0; i < width; i++)
        img->imgY[j][i] = (unsigned short) *tmp_buf++;

    for(uv=0; uv < 2; uv++)
      for(j=0; j < height/2 ; j++)
        for(i=0; i < width/2; i++)
          img->imgUV[uv][j][i] = (unsigned short) *tmp_buf++;
  }
  else
  {
    printf("Error reading frame\n");
  }

  return (status);
}

//
//
//
void write_image(int height, int width, FILE *p_out, imgStruct img)
{
  int i, j, k;
  int frameSize = height * width * 3 / 2;
  static byte *buf = NULL;
  byte *tmp_buf = NULL;
  static int first = 1;

  if(first)
  {
    buf = (byte *) malloc(frameSize * sizeof(byte));
    first = 0;
  }
    
  tmp_buf = buf;

  for(i=0; i < height; i++)
    for(j=0; j < width; j++)
      *tmp_buf++ = (byte) img.imgY[i][j];
    
  for(k=0; k < 2; ++k)
    for(i=0; i < height/2; i++)
      for(j=0; j < width/2; j++)
      *tmp_buf++ = (byte) img.imgUV[k][i][j];

  fwrite(buf, frameSize * sizeof(byte), 1, p_out);
}


//
//
//
Bitstream * allocBitstream(int size)
{
  Bitstream *tmp;
  tmp = (Bitstream *) calloc(1, sizeof(Bitstream));
  tmp->streamBuffer = (byte *) calloc(size, sizeof(byte));
  tmp->frame_bitoffset = 0;
  tmp->bitstream_length = 0;
  return tmp;
}

//
//
//
void freeBitstream(Bitstream *bitstream)
{
  free(bitstream->streamBuffer);
  bitstream->streamBuffer = NULL;
  free(bitstream);
}

//
//
//
int readBitstream(Bitstream *bitstream, char *fname)
{
  FILE *fp;
  int flen = 0;
  if((fp = fopen(fname, "rb")) == 0)
  {
    printf("Error open file %s\n", fname);
    exit(0);
  } 
  fseek(fp, 0L, SEEK_END);
  flen = (int) ftell(fp);
  fseek(fp, 0L, SEEK_SET);
  fread(bitstream->streamBuffer, sizeof(byte), flen, fp);
  fclose(fp);
  bitstream->bitstream_length = flen;
  bitstream->frame_bitoffset = 0;
  printf("Read %d bytes from file %s\n", flen, fname);
  return flen;
}

//
//
//
int GetBits (byte buffer[], int totbitoffset, int *info, int bytecount, int numbits)
{
  register int inf;
  long byteoffset;    // byte from start of buffer
  int bitoffset;      // bit from start of byte

  int bitcounter = numbits;

  byteoffset = totbitoffset / 8;
  bitoffset = 7 - (totbitoffset % 8);

  inf = 0;
  while(numbits)
  {
    inf <<=1;
    inf |= (buffer[byteoffset] & (0x01 << bitoffset)) >> bitoffset;
    numbits--;
    bitoffset--;
    if(bitoffset < 0)
    {
      byteoffset++;
      bitoffset += 8;
      if(byteoffset > bytecount)
        return -1;
    }
  }

  *info = inf;
  return bitcounter;  // return absolute offset in bit from start of frame
}     


//
//
//
int readSyntaxElement_FLC(SyntaxElement *sym, Bitstream *currStream)
{
  int frame_bitoffset = currStream->frame_bitoffset;
  byte *buf = currStream->streamBuffer;
  int BitstreamLengthInBytes = currStream->bitstream_length;

  if ((GetBits(buf, frame_bitoffset, &(sym->inf), BitstreamLengthInBytes, sym->len)) < 0)
    return -1;

  currStream->frame_bitoffset += sym->len; // move bitstream pointer
  sym->value1 = sym->inf;

  return 1;
}

//
//
//
void linfo_ue(int len, int info, int *value1, int *dummy)
{
  assert(len / 2 < 32);
  *value1 = (1 << (len / 2)) + info - 1;
}

//
//
//
int u_1(char *tracestring, Bitstream *bitstream)
{
  return u_v(1, tracestring, bitstream);
}

//
//
//
int u_v(int LenInBits, char *tracestring, Bitstream *bitstream)
{
  SyntaxElement symbol, *sym = &symbol;
  symbol.inf = 0;//required by the compiler
  assert (bitstream->streamBuffer != NULL);
  sym->type = 0; // SE_HEADER;
  sym->mapping = linfo_ue;   // Mapping rule
  sym->len = LenInBits;
  readSyntaxElement_FLC(sym, bitstream);
  return sym->inf;
};

//
// from http://en.wikipedia.org/wiki/Golomb_coding
// M=2^k
//
int golombDecode(int k, Bitstream *bitstream, char *tracestring)
{
  int q = 0;
  int nr = 0;
  int m = (int)pow(2.0, k);
  int a;

  while(u_1("full-pel filter", bitstream)) 
    q++;
  for(a = 0; a < k; ++a)          // read out the sequential log2(M) bits
    if(u_1("full-pel filter", bitstream))
      nr += 1 << a;
  nr += q * m;                    // add the bits and the multiple of M
  if(nr != 0)
    nr = (u_1("full-pel filter", bitstream))? nr: -nr;
  return nr; 
}

//
//  force0: 
//    0 - normal
//    1 - force0 flag followed by codedVarBins[0...filters_per_group-1]
//
//  predMethod:
//    0 - filter coefficients non predicted
//    1 - filter coefficients predicted
//
void readPredFlags(int *force0, int *predMethod, int codedVarBins[NO_VAR_BINS], 
                   int filters_per_group, Bitstream *bitstream)
{
  int ind;

  *force0 = u_1("force0", bitstream);
  if(*force0)
    for(ind = 0; ind < filters_per_group; ++ind)
      codedVarBins[ind] = u_1("codedVarBins[i]", bitstream);
  *predMethod = u_1("predMethod", bitstream);
}


//
//
//
void readFilterCodingParams(int *minKStart, int kMinTab[], int sqrFiltLength, Bitstream *bitstream)
{
  int ind, scanPos;
  int golombIndexBit;
  int kMin;
  int maxScanVal;
  int *pDepthInt;
  int fl;

  // Determine fl
  if(sqrFiltLength == SQR_FILT_LENGTH_9SYM)
    fl = 4;
  else if(sqrFiltLength == SQR_FILT_LENGTH_7SYM)
    fl = 3;
  else
    fl = 2;

  // Determine maxScanVal
  maxScanVal = 0;
  pDepthInt = pDepthIntTab[fl - 2];
  for(ind = 0; ind < sqrFiltLength; ind++)
    maxScanVal = max(maxScanVal, pDepthInt[ind]);

  // Golomb parameters
  *minKStart = 1 + u_v(3, "expGolomb index", bitstream);

  kMin = *minKStart;
  for(scanPos = 0; scanPos < maxScanVal; scanPos++)
  {
    golombIndexBit = u_1("expGolomb index", bitstream);
    if(golombIndexBit)
      kMinTab[scanPos] = kMin + 1;
    else
      kMinTab[scanPos] = kMin;
    kMin = kMinTab[scanPos];
  }
}


//
//
//
void readFilterCoeffs(int sqrFiltLength, int savedFilters, int **FilterCoeff, int kMinTab[], 
                      Bitstream *bitstream)
{
  int ind, scanPos, i;
  int *pDepthInt;
  int fl;

  if(sqrFiltLength == SQR_FILT_LENGTH_9SYM)
    fl = 4;
  else if(sqrFiltLength == SQR_FILT_LENGTH_7SYM)
    fl = 3;
  else
    fl = 2;
    
  pDepthInt = pDepthIntTab[fl - 2];

  for(ind = 0; ind < savedFilters; ++ind)
  {
    for(i = 0; i < sqrFiltLength; i++)
    {	
      scanPos = pDepthInt[i] - 1;
      FilterCoeff[ind][i] = golombDecode(kMinTab[scanPos], bitstream, "full-pel filter");
    }
  }
}

//
//
//
void decodeAuxInfo(int *nonZeroCoeffs, int *sqrFiltLength, int *filters_per_group, 
                   int varIndTab[NO_VAR_BINS], int *prevFiltInd, int usePrevFiltBest[NO_VAR_BINS], 
                   int frNo, Bitstream *bitstream)
{
  int i;
  int filterPattern[NO_VAR_BINS] = {0};
  int predPattern;
  int tmp;
  int sqrFiltLengthTab[] = {SQR_FILT_LENGTH_9SYM, SQR_FILT_LENGTH_7SYM, SQR_FILT_LENGTH_5SYM};

  *nonZeroCoeffs = 1;
  
  if(frNo > 0)
    *nonZeroCoeffs = u_1("Nonzero coefficients", bitstream);

  if(*nonZeroCoeffs)
  {
    //  filtNo decoding:
    //    00   *sqrFiltLength =  SQR_FILT_LENGTH_9SYM
    //    01   *sqrFiltLength =  SQR_FILT_LENGTH_7SYM
    //    1    *sqrFiltLength =  SQR_FILT_LENGTH_5SYM
    //
    tmp = u_1("codtab_filtNo", bitstream);
    if(tmp)
    {
      *sqrFiltLength = sqrFiltLengthTab[2];
    }
    else
    {
      tmp += u_1("codtab_filtNo", bitstream);
      *sqrFiltLength = sqrFiltLengthTab[tmp];
    }

    //  noFilters decoding:
    //    1     *filters_per_group = 1
    //    01    *filters_per_group = 2
    //    00    *filters_per_group = more than 2 (exact number from filterPattern[])
    //
    tmp = u_1("codtab_noFilters", bitstream);
    if(tmp)
    {
      *filters_per_group = 1;
    }
    else
    {
      tmp = u_1("codtab_noFilters", bitstream);
      if(tmp)
        *filters_per_group = 2;
      else
        *filters_per_group = 3;    // More than two filters
    }

    if(*filters_per_group == 2)
    {
      int startSecondFilter = u_v (4, "startSecondFilter", bitstream);
      filterPattern[startSecondFilter] = 1;
    }
    else if(*filters_per_group > 2)
    {
      *filters_per_group = 1;
      for(i = 1; i < NO_VAR_BINS; ++i)
      {
        filterPattern[i] = u_1("filterPattern[i]", bitstream);
        *filters_per_group += filterPattern[i];
      }
    }
    else
    {
      memset(filterPattern, 0, NO_VAR_BINS * sizeof(int));
    }
  }

  // Reconstruct varIndTab[]
  memset(varIndTab, 0, NO_VAR_BINS * sizeof(int));
  if(*nonZeroCoeffs)
  {
    for(i = 1; i < NO_VAR_BINS; ++i)
      if(filterPattern[i])
        varIndTab[i] = varIndTab[i-1] + 1;
      else
        varIndTab[i] = varIndTab[i-1];
  }

  if(frNo > 0)
  {
    // Frame that filter is predicted from
    *prevFiltInd = u_v(4, "prevFiltInd", bitstream);

    // 
    //  prediction pattern:
    //    1     predPattern = 0   --> usePrevFiltBest[] = "1111111111111111"
    //    01    predPattern = 1   --> usePrevFiltBest[] = "0111111111111111"
    //    00    predPattern = 2   --> usePrevFiltBest[] = actual pattern sent
    //
    tmp = u_1("codtab_pred", bitstream);
    if(tmp)
    {
      predPattern = 0;
    }
    else
    {
      tmp = u_1("codtab_pred", bitstream);
      if(tmp)
        predPattern = 1;
      else
        predPattern = 2;
    }

    // Reconstruct usePrevFiltBest[]
    if(predPattern == 0)
    {
      for(i = 0; i < NO_VAR_BINS; ++i)
        usePrevFiltBest[i] = 1;
    } 
    else if(predPattern == 1)
    {
      usePrevFiltBest[0] = 0;
      for(i = 1; i < NO_VAR_BINS; ++i)
        usePrevFiltBest[i] = 1;
    }
    else
    {
      for(i = 0; i < NO_VAR_BINS; ++i)
        usePrevFiltBest[i] = u_1("usePrevFiltBest[i]", bitstream);
    }
  }
}


